# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/fmpz_poly.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpz_poly_init(fmpz_poly_t poly) noexcept
    void fmpz_poly_init2(fmpz_poly_t poly, slong alloc) noexcept
    void fmpz_poly_realloc(fmpz_poly_t poly, slong alloc) noexcept
    void fmpz_poly_fit_length(fmpz_poly_t poly, slong len) noexcept
    void fmpz_poly_clear(fmpz_poly_t poly) noexcept
    void _fmpz_poly_normalise(fmpz_poly_t poly) noexcept
    void _fmpz_poly_set_length(fmpz_poly_t poly, slong newlen) noexcept
    void fmpz_poly_attach_truncate(fmpz_poly_t trunc, const fmpz_poly_t poly, slong n) noexcept
    void fmpz_poly_attach_shift(fmpz_poly_t trunc, const fmpz_poly_t poly, slong n) noexcept
    slong fmpz_poly_length(const fmpz_poly_t poly) noexcept
    slong fmpz_poly_degree(const fmpz_poly_t poly) noexcept
    void fmpz_poly_set(fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_set_si(fmpz_poly_t poly, slong c) noexcept
    void fmpz_poly_set_ui(fmpz_poly_t poly, ulong c) noexcept
    void fmpz_poly_set_fmpz(fmpz_poly_t poly, const fmpz_t c) noexcept
    int _fmpz_poly_set_str(fmpz * poly, const char * str) noexcept
    int fmpz_poly_set_str(fmpz_poly_t poly, const char * str) noexcept
    char * _fmpz_poly_get_str(const fmpz * poly, slong len) noexcept
    char * fmpz_poly_get_str(const fmpz_poly_t poly) noexcept
    char * _fmpz_poly_get_str_pretty(const fmpz * poly, slong len, const char * x) noexcept
    char * fmpz_poly_get_str_pretty(const fmpz_poly_t poly, const char * x) noexcept
    void fmpz_poly_zero(fmpz_poly_t poly) noexcept
    void fmpz_poly_one(fmpz_poly_t poly) noexcept
    void fmpz_poly_zero_coeffs(fmpz_poly_t poly, slong i, slong j) noexcept
    void fmpz_poly_swap(fmpz_poly_t poly1, fmpz_poly_t poly2) noexcept
    void _fmpz_poly_reverse(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_reverse(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void fmpz_poly_truncate(fmpz_poly_t poly, slong newlen) noexcept
    void fmpz_poly_set_trunc(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void fmpz_poly_randtest(fmpz_poly_t f, flint_rand_t state, slong len, flint_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_unsigned(fmpz_poly_t f, flint_rand_t state, slong len, flint_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_not_zero(fmpz_poly_t f, flint_rand_t state, slong len, flint_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_no_real_root(fmpz_poly_t p, flint_rand_t state, slong len, flint_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_irreducible1(fmpz_poly_t pol, flint_rand_t state, slong len, mp_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_irreducible2(fmpz_poly_t pol, flint_rand_t state, slong len, mp_bitcnt_t bits) noexcept
    void fmpz_poly_randtest_irreducible(fmpz_poly_t pol, flint_rand_t state, slong len, mp_bitcnt_t bits) noexcept
    void fmpz_poly_get_coeff_fmpz(fmpz_t x, const fmpz_poly_t poly, slong n) noexcept
    slong fmpz_poly_get_coeff_si(const fmpz_poly_t poly, slong n) noexcept
    ulong fmpz_poly_get_coeff_ui(const fmpz_poly_t poly, slong n) noexcept
    fmpz * fmpz_poly_get_coeff_ptr(const fmpz_poly_t poly, slong n) noexcept
    fmpz * fmpz_poly_lead(const fmpz_poly_t poly) noexcept
    void fmpz_poly_set_coeff_fmpz(fmpz_poly_t poly, slong n, const fmpz_t x) noexcept
    void fmpz_poly_set_coeff_si(fmpz_poly_t poly, slong n, slong x) noexcept
    void fmpz_poly_set_coeff_ui(fmpz_poly_t poly, slong n, ulong x) noexcept
    bint fmpz_poly_equal(const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    bint fmpz_poly_equal_trunc(const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    bint fmpz_poly_is_zero(const fmpz_poly_t poly) noexcept
    bint fmpz_poly_is_one(const fmpz_poly_t poly) noexcept
    bint fmpz_poly_is_unit(const fmpz_poly_t poly) noexcept
    bint fmpz_poly_is_gen(const fmpz_poly_t poly) noexcept
    void _fmpz_poly_add(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_add(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_add_series(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_sub(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_sub(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_sub_series(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void fmpz_poly_neg(fmpz_poly_t res, const fmpz_poly_t poly) noexcept
    void fmpz_poly_scalar_abs(fmpz_poly_t res, const fmpz_poly_t poly) noexcept
    void fmpz_poly_scalar_mul_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_mul_si(fmpz_poly_t poly1, const fmpz_poly_t poly2, slong x) noexcept
    void fmpz_poly_scalar_mul_ui(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_mul_2exp(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong exp) noexcept
    void fmpz_poly_scalar_addmul_si(fmpz_poly_t poly1, const fmpz_poly_t poly2, slong x) noexcept
    void fmpz_poly_scalar_addmul_ui(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_addmul_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_submul_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_fdiv_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_fdiv_si(fmpz_poly_t poly1, const fmpz_poly_t poly2, slong x) noexcept
    void fmpz_poly_scalar_fdiv_ui(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_fdiv_2exp(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_tdiv_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_tdiv_si(fmpz_poly_t poly1, const fmpz_poly_t poly2, slong x) noexcept
    void fmpz_poly_scalar_tdiv_ui(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_tdiv_2exp(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_divexact_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t x) noexcept
    void fmpz_poly_scalar_divexact_si(fmpz_poly_t poly1, const fmpz_poly_t poly2, slong x) noexcept
    void fmpz_poly_scalar_divexact_ui(fmpz_poly_t poly1, const fmpz_poly_t poly2, ulong x) noexcept
    void fmpz_poly_scalar_mod_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t p) noexcept
    void fmpz_poly_scalar_smod_fmpz(fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t p) noexcept
    slong _fmpz_poly_remove_content_2exp(fmpz * pol, slong len) noexcept
    void _fmpz_poly_scale_2exp(fmpz * pol, slong len, slong k) noexcept
    void _fmpz_poly_bit_pack(mp_ptr arr, const fmpz * poly, slong len, flint_bitcnt_t bit_size, int negate) noexcept
    int _fmpz_poly_bit_unpack(fmpz * poly, slong len, mp_srcptr arr, flint_bitcnt_t bit_size, int negate) noexcept
    void _fmpz_poly_bit_unpack_unsigned(fmpz * poly, slong len, mp_srcptr arr, flint_bitcnt_t bit_size) noexcept
    void fmpz_poly_bit_pack(fmpz_t f, const fmpz_poly_t poly, flint_bitcnt_t bit_size) noexcept
    void fmpz_poly_bit_unpack(fmpz_poly_t poly, const fmpz_t f, flint_bitcnt_t bit_size) noexcept
    void fmpz_poly_bit_unpack_unsigned(fmpz_poly_t poly, const fmpz_t f, flint_bitcnt_t bit_size) noexcept
    void _fmpz_poly_mul_classical(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_mul_classical(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mullow_classical(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_mullow_classical(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_mulhigh_classical(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong start) noexcept
    void fmpz_poly_mulhigh_classical(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong start) noexcept
    void _fmpz_poly_mulmid_classical(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_mulmid_classical(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mul_karatsuba(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_mul_karatsuba(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mullow_karatsuba_n(fmpz * res, const fmpz * poly1, const fmpz * poly2, slong n) noexcept
    void fmpz_poly_mullow_karatsuba_n(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_mulhigh_karatsuba_n(fmpz * res, const fmpz * poly1, const fmpz * poly2, slong len) noexcept
    void fmpz_poly_mulhigh_karatsuba_n(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong len) noexcept
    void _fmpz_poly_mul_KS(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_mul_KS(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mullow_KS(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_mullow_KS(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_mul_SS(fmpz * output, const fmpz * input1, slong length1, const fmpz * input2, slong length2) noexcept
    void fmpz_poly_mul_SS(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mullow_SS(fmpz * output, const fmpz * input1, slong length1, const fmpz * input2, slong length2, slong n) noexcept
    void fmpz_poly_mullow_SS(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_mul(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_mul(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_mullow(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_mullow(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void fmpz_poly_mulhigh_n(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_mulhigh(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong start) noexcept
    void fmpz_poly_mul_SS_precache_init(fmpz_poly_mul_precache_t pre, slong len1, slong bits1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_mul_precache_clear(fmpz_poly_mul_precache_t pre) noexcept
    void _fmpz_poly_mullow_SS_precache(fmpz * output, const fmpz * input1, slong len1, fmpz_poly_mul_precache_t pre, slong trunc) noexcept
    void fmpz_poly_mullow_SS_precache(fmpz_poly_t res, const fmpz_poly_t poly1, fmpz_poly_mul_precache_t pre, slong n) noexcept
    void fmpz_poly_mul_SS_precache(fmpz_poly_t res, const fmpz_poly_t poly1, fmpz_poly_mul_precache_t pre) noexcept
    void _fmpz_poly_sqr_KS(fmpz * rop, const fmpz * op, slong len) noexcept
    void fmpz_poly_sqr_KS(fmpz_poly_t rop, const fmpz_poly_t op) noexcept
    void _fmpz_poly_sqr_karatsuba(fmpz * rop, const fmpz * op, slong len) noexcept
    void fmpz_poly_sqr_karatsuba(fmpz_poly_t rop, const fmpz_poly_t op) noexcept
    void _fmpz_poly_sqr_classical(fmpz * rop, const fmpz * op, slong len) noexcept
    void fmpz_poly_sqr_classical(fmpz_poly_t rop, const fmpz_poly_t op) noexcept
    void _fmpz_poly_sqr(fmpz * rop, const fmpz * op, slong len) noexcept
    void fmpz_poly_sqr(fmpz_poly_t rop, const fmpz_poly_t op) noexcept
    void _fmpz_poly_sqrlow_KS(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_sqrlow_KS(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_sqrlow_karatsuba_n(fmpz * res, const fmpz * poly, slong n) noexcept
    void fmpz_poly_sqrlow_karatsuba_n(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_sqrlow_classical(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_sqrlow_classical(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_sqrlow(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_sqrlow(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_pow_multinomial(fmpz * res, const fmpz * poly, slong len, ulong e) noexcept
    void fmpz_poly_pow_multinomial(fmpz_poly_t res, const fmpz_poly_t poly, ulong e) noexcept
    void _fmpz_poly_pow_binomial(fmpz * res, const fmpz * poly, ulong e) noexcept
    void fmpz_poly_pow_binomial(fmpz_poly_t res, const fmpz_poly_t poly, ulong e) noexcept
    void _fmpz_poly_pow_addchains(fmpz * res, const fmpz * poly, slong len, const int * a, int n) noexcept
    void fmpz_poly_pow_addchains(fmpz_poly_t res, const fmpz_poly_t poly, ulong e) noexcept
    void _fmpz_poly_pow_binexp(fmpz * res, const fmpz * poly, slong len, ulong e) noexcept
    void fmpz_poly_pow_binexp(fmpz_poly_t res, const fmpz_poly_t poly, ulong e) noexcept
    void _fmpz_poly_pow_small(fmpz * res, const fmpz * poly, slong len, ulong e) noexcept
    void _fmpz_poly_pow(fmpz * res, const fmpz * poly, slong len, ulong e) noexcept
    void fmpz_poly_pow(fmpz_poly_t res, const fmpz_poly_t poly, ulong e) noexcept
    void _fmpz_poly_pow_trunc(fmpz * res, const fmpz * poly, ulong e, slong n) noexcept
    void fmpz_poly_pow_trunc(fmpz_poly_t res, const fmpz_poly_t poly, ulong e, slong n) noexcept
    void _fmpz_poly_shift_left(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_shift_left(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_shift_right(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_shift_right(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    ulong fmpz_poly_max_limbs(const fmpz_poly_t poly) noexcept
    slong fmpz_poly_max_bits(const fmpz_poly_t poly) noexcept
    void fmpz_poly_height(fmpz_t height, const fmpz_poly_t poly) noexcept
    void _fmpz_poly_2norm(fmpz_t res, const fmpz * poly, slong len) noexcept
    void fmpz_poly_2norm(fmpz_t res, const fmpz_poly_t poly) noexcept
    mp_limb_t _fmpz_poly_2norm_normalised_bits(const fmpz * poly, slong len) noexcept
    void _fmpz_poly_gcd_subresultant(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_gcd_subresultant(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    int _fmpz_poly_gcd_heuristic(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    int fmpz_poly_gcd_heuristic(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_gcd_modular(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_gcd_modular(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_gcd(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_gcd(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_xgcd_modular(fmpz_t r, fmpz * s, fmpz * t, const fmpz * f, slong len1, const fmpz * g, slong len2) noexcept
    void fmpz_poly_xgcd_modular(fmpz_t r, fmpz_poly_t s, fmpz_poly_t t, const fmpz_poly_t f, const fmpz_poly_t g) noexcept
    void _fmpz_poly_xgcd(fmpz_t r, fmpz * s, fmpz * t, const fmpz * f, slong len1, const fmpz * g, slong len2) noexcept
    void fmpz_poly_xgcd(fmpz_t r, fmpz_poly_t s, fmpz_poly_t t, const fmpz_poly_t f, const fmpz_poly_t g) noexcept
    void _fmpz_poly_lcm(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_lcm(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_resultant_modular(fmpz_t res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_resultant_modular(fmpz_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_resultant_modular_div(fmpz_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, const fmpz_t div, slong nbits) noexcept
    void _fmpz_poly_resultant_euclidean(fmpz_t res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_resultant_euclidean(fmpz_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_resultant(fmpz_t res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_resultant(fmpz_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_discriminant(fmpz_t res, const fmpz * poly, slong len) noexcept
    void fmpz_poly_discriminant(fmpz_t res, const fmpz_poly_t poly) noexcept
    void _fmpz_poly_content(fmpz_t res, const fmpz * poly, slong len) noexcept
    void fmpz_poly_content(fmpz_t res, const fmpz_poly_t poly) noexcept
    void _fmpz_poly_primitive_part(fmpz * res, const fmpz * poly, slong len) noexcept
    void fmpz_poly_primitive_part(fmpz_poly_t res, const fmpz_poly_t poly) noexcept
    bint _fmpz_poly_is_squarefree(const fmpz * poly, slong len) noexcept
    bint fmpz_poly_is_squarefree(const fmpz_poly_t poly) noexcept
    int _fmpz_poly_divrem_basecase(fmpz * Q, fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_divrem_basecase(fmpz_poly_t Q, fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    int _fmpz_poly_divrem_divconquer_recursive(fmpz * Q, fmpz * BQ, fmpz * W, const fmpz * A, const fmpz * B, slong lenB, int exact) noexcept
    int _fmpz_poly_divrem_divconquer(fmpz * Q, fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_divrem_divconquer(fmpz_poly_t Q, fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    int _fmpz_poly_divrem(fmpz * Q, fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_divrem(fmpz_poly_t Q, fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    int _fmpz_poly_div_basecase(fmpz * Q, fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_div_basecase(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    int _fmpz_poly_divremlow_divconquer_recursive(fmpz * Q, fmpz * BQ, const fmpz * A, const fmpz * B, slong lenB, int exact) noexcept
    int _fmpz_poly_div_divconquer_recursive(fmpz * Q, fmpz * temp, const fmpz * A, const fmpz * B, slong lenB, int exact) noexcept
    int _fmpz_poly_div_divconquer(fmpz * Q, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_div_divconquer(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    int _fmpz_poly_div(fmpz * Q, const fmpz * A, slong lenA, const fmpz * B, slong lenB, int exact) noexcept
    void fmpz_poly_div(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_rem_basecase(fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB) noexcept
    void fmpz_poly_rem_basecase(fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_rem(fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB) noexcept
    void fmpz_poly_rem(fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_div_root(fmpz * Q, const fmpz * A, slong len, const fmpz_t c) noexcept
    void fmpz_poly_div_root(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_t c) noexcept
    void _fmpz_poly_preinvert(fmpz * B_inv, const fmpz * B, slong n) noexcept
    void fmpz_poly_preinvert(fmpz_poly_t B_inv, const fmpz_poly_t B) noexcept
    void _fmpz_poly_div_preinv(fmpz * Q, const fmpz * A, slong len1, const fmpz * B, const fmpz * B_inv, slong len2) noexcept
    void fmpz_poly_div_preinv(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B, const fmpz_poly_t B_inv) noexcept
    void _fmpz_poly_divrem_preinv(fmpz * Q, fmpz * A, slong len1, const fmpz * B, const fmpz * B_inv, slong len2) noexcept
    void fmpz_poly_divrem_preinv(fmpz_poly_t Q, fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B, const fmpz_poly_t B_inv) noexcept
    fmpz ** _fmpz_poly_powers_precompute(const fmpz * B, slong len) noexcept
    void fmpz_poly_powers_precompute(fmpz_poly_powers_precomp_t pinv, fmpz_poly_t poly) noexcept
    void _fmpz_poly_powers_clear(fmpz ** powers, slong len) noexcept
    void fmpz_poly_powers_clear(fmpz_poly_powers_precomp_t pinv) noexcept
    void _fmpz_poly_rem_powers_precomp(fmpz * A, slong m, const fmpz * B, slong n, fmpz ** const powers) noexcept
    void fmpz_poly_rem_powers_precomp(fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B, const fmpz_poly_powers_precomp_t B_inv) noexcept
    int _fmpz_poly_divides(fmpz * Q, const fmpz * A, slong lenA, const fmpz * B, slong lenB) noexcept
    int fmpz_poly_divides(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    slong fmpz_poly_remove(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_divlow_smodp(fmpz * res, const fmpz_poly_t f, const fmpz_poly_t g, const fmpz_t p, slong n) noexcept
    void fmpz_poly_divhigh_smodp(fmpz * res, const fmpz_poly_t f, const fmpz_poly_t g, const fmpz_t p, slong n) noexcept
    void _fmpz_poly_inv_series_basecase(fmpz * Qinv, const fmpz * Q, slong Qlen, slong n) noexcept
    void fmpz_poly_inv_series_basecase(fmpz_poly_t Qinv, const fmpz_poly_t Q, slong n) noexcept
    void _fmpz_poly_inv_series_newton(fmpz * Qinv, const fmpz * Q, slong Qlen, slong n) noexcept
    void fmpz_poly_inv_series_newton(fmpz_poly_t Qinv, const fmpz_poly_t Q, slong n) noexcept
    void _fmpz_poly_inv_series(fmpz * Qinv, const fmpz * Q, slong Qlen, slong n) noexcept
    void fmpz_poly_inv_series(fmpz_poly_t Qinv, const fmpz_poly_t Q, slong n) noexcept
    void _fmpz_poly_div_series_basecase(fmpz * Q, const fmpz * A, slong Alen, const fmpz * B, slong Blen, slong n) noexcept
    void _fmpz_poly_div_series_divconquer(fmpz * Q, const fmpz * A, slong Alen, const fmpz * B, slong Blen, slong n) noexcept
    void _fmpz_poly_div_series(fmpz * Q, const fmpz * A, slong Alen, const fmpz * B, slong Blen, slong n) noexcept
    void fmpz_poly_div_series_basecase(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B, slong n) noexcept
    void fmpz_poly_div_series_divconquer(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B, slong n) noexcept
    void fmpz_poly_div_series(fmpz_poly_t Q, const fmpz_poly_t A, const fmpz_poly_t B, slong n) noexcept
    void _fmpz_poly_pseudo_divrem_basecase(fmpz * Q, fmpz * R, ulong * d, const fmpz * A, slong lenA, const fmpz * B, slong lenB, const fmpz_preinvn_t inv) noexcept
    void fmpz_poly_pseudo_divrem_basecase(fmpz_poly_t Q, fmpz_poly_t R, ulong * d, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_divrem_divconquer(fmpz * Q, fmpz * R, ulong * d, const fmpz * A, slong lenA, const fmpz * B, slong lenB, const fmpz_preinvn_t inv) noexcept
    void fmpz_poly_pseudo_divrem_divconquer(fmpz_poly_t Q, fmpz_poly_t R, ulong * d, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_divrem_cohen(fmpz * Q, fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB) noexcept
    void fmpz_poly_pseudo_divrem_cohen(fmpz_poly_t Q, fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_rem_cohen(fmpz * R, const fmpz * A, slong lenA, const fmpz * B, slong lenB) noexcept
    void fmpz_poly_pseudo_rem_cohen(fmpz_poly_t R, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_divrem(fmpz * Q, fmpz * R, ulong * d, const fmpz * A, slong lenA, const fmpz * B, slong lenB, const fmpz_preinvn_t inv) noexcept
    void fmpz_poly_pseudo_divrem(fmpz_poly_t Q, fmpz_poly_t R, ulong * d, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_div(fmpz * Q, ulong * d, const fmpz * A, slong lenA, const fmpz * B, slong lenB, const fmpz_preinvn_t inv) noexcept
    void fmpz_poly_pseudo_div(fmpz_poly_t Q, ulong * d, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_pseudo_rem(fmpz * R, ulong * d, const fmpz * A, slong lenA, const fmpz * B, slong lenB, const fmpz_preinvn_t inv) noexcept
    void fmpz_poly_pseudo_rem(fmpz_poly_t R, ulong * d, const fmpz_poly_t A, const fmpz_poly_t B) noexcept
    void _fmpz_poly_derivative(fmpz * rpoly, const fmpz * poly, slong len) noexcept
    void fmpz_poly_derivative(fmpz_poly_t res, const fmpz_poly_t poly) noexcept
    void _fmpz_poly_nth_derivative(fmpz * rpoly, const fmpz * poly, ulong n, slong len) noexcept
    void fmpz_poly_nth_derivative(fmpz_poly_t res, const fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_evaluate_divconquer_fmpz(fmpz_t res, const fmpz * poly, slong len, const fmpz_t a) noexcept
    void fmpz_poly_evaluate_divconquer_fmpz(fmpz_t res, const fmpz_poly_t poly, const fmpz_t a) noexcept
    void _fmpz_poly_evaluate_horner_fmpz(fmpz_t res, const fmpz * f, slong len, const fmpz_t a) noexcept
    void fmpz_poly_evaluate_horner_fmpz(fmpz_t res, const fmpz_poly_t f, const fmpz_t a) noexcept
    void _fmpz_poly_evaluate_fmpz(fmpz_t res, const fmpz * f, slong len, const fmpz_t a) noexcept
    void fmpz_poly_evaluate_fmpz(fmpz_t res, const fmpz_poly_t f, const fmpz_t a) noexcept
    void _fmpz_poly_evaluate_divconquer_fmpq(fmpz_t rnum, fmpz_t rden, const fmpz * f, slong len, const fmpz_t anum, const fmpz_t aden) noexcept
    void fmpz_poly_evaluate_divconquer_fmpq(fmpq_t res, const fmpz_poly_t f, const fmpq_t a) noexcept
    void _fmpz_poly_evaluate_horner_fmpq(fmpz_t rnum, fmpz_t rden, const fmpz * f, slong len, const fmpz_t anum, const fmpz_t aden) noexcept
    void fmpz_poly_evaluate_horner_fmpq(fmpq_t res, const fmpz_poly_t f, const fmpq_t a) noexcept
    void _fmpz_poly_evaluate_fmpq(fmpz_t rnum, fmpz_t rden, const fmpz * f, slong len, const fmpz_t anum, const fmpz_t aden) noexcept
    void fmpz_poly_evaluate_fmpq(fmpq_t res, const fmpz_poly_t f, const fmpq_t a) noexcept
    mp_limb_t _fmpz_poly_evaluate_mod(const fmpz * poly, slong len, mp_limb_t a, mp_limb_t n, mp_limb_t ninv) noexcept
    mp_limb_t fmpz_poly_evaluate_mod(const fmpz_poly_t poly, mp_limb_t a, mp_limb_t n) noexcept
    void fmpz_poly_evaluate_fmpz_vec(fmpz * res, const fmpz_poly_t f, const fmpz * a, slong n) noexcept
    double _fmpz_poly_evaluate_horner_d(const fmpz * poly, slong n, double d) noexcept
    double fmpz_poly_evaluate_horner_d(const fmpz_poly_t poly, double d) noexcept
    double _fmpz_poly_evaluate_horner_d_2exp(slong * exp, const fmpz * poly, slong n, double d) noexcept
    double fmpz_poly_evaluate_horner_d_2exp(slong * exp, const fmpz_poly_t poly, double d) noexcept
    double _fmpz_poly_evaluate_horner_d_2exp2(slong * exp, const fmpz * poly, slong n, double d, slong dexp) noexcept
    void _fmpz_poly_monomial_to_newton(fmpz * poly, const fmpz * roots, slong n) noexcept
    void _fmpz_poly_newton_to_monomial(fmpz * poly, const fmpz * roots, slong n) noexcept
    void fmpz_poly_interpolate_fmpz_vec(fmpz_poly_t poly, const fmpz * xs, const fmpz * ys, slong n) noexcept
    void _fmpz_poly_compose_horner(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_compose_horner(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_compose_divconquer(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_compose_divconquer(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void _fmpz_poly_compose(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2) noexcept
    void fmpz_poly_compose(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2) noexcept
    void fmpz_poly_inflate(fmpz_poly_t result, const fmpz_poly_t input, ulong inflation) noexcept
    void fmpz_poly_deflate(fmpz_poly_t result, const fmpz_poly_t input, ulong deflation) noexcept
    ulong fmpz_poly_deflation(const fmpz_poly_t input) noexcept
    void _fmpz_poly_taylor_shift_horner(fmpz * poly, const fmpz_t c, slong n) noexcept
    void fmpz_poly_taylor_shift_horner(fmpz_poly_t g, const fmpz_poly_t f, const fmpz_t c) noexcept
    void _fmpz_poly_taylor_shift_divconquer(fmpz * poly, const fmpz_t c, slong n) noexcept
    void fmpz_poly_taylor_shift_divconquer(fmpz_poly_t g, const fmpz_poly_t f, const fmpz_t c) noexcept
    void _fmpz_poly_taylor_shift_multi_mod(fmpz * poly, const fmpz_t c, slong n) noexcept
    void fmpz_poly_taylor_shift_multi_mod(fmpz_poly_t g, const fmpz_poly_t f, const fmpz_t c) noexcept
    void _fmpz_poly_taylor_shift(fmpz * poly, const fmpz_t c, slong n) noexcept
    void fmpz_poly_taylor_shift(fmpz_poly_t g, const fmpz_poly_t f, const fmpz_t c) noexcept
    void _fmpz_poly_compose_series_horner(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_compose_series_horner(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_compose_series_brent_kung(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_compose_series_brent_kung(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_compose_series(fmpz * res, const fmpz * poly1, slong len1, const fmpz * poly2, slong len2, slong n) noexcept
    void fmpz_poly_compose_series(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_poly_t poly2, slong n) noexcept
    void _fmpz_poly_revert_series(fmpz * Qinv, const fmpz * Q, slong Qlen, slong n) noexcept
    void fmpz_poly_revert_series(fmpz_poly_t Qinv, const fmpz_poly_t Q, slong n) noexcept
    int _fmpz_poly_sqrtrem_classical(fmpz * res, fmpz * r, const fmpz * poly, slong len) noexcept
    int fmpz_poly_sqrtrem_classical(fmpz_poly_t b, fmpz_poly_t r, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrtrem_divconquer(fmpz * res, fmpz * r, const fmpz * poly, slong len, fmpz * temp) noexcept
    int fmpz_poly_sqrtrem_divconquer(fmpz_poly_t b, fmpz_poly_t r, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrt_classical(fmpz * res, const fmpz * poly, slong len, int exact) noexcept
    int fmpz_poly_sqrt_classical(fmpz_poly_t b, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrt_KS(fmpz * res, const fmpz * poly, slong len) noexcept
    int fmpz_poly_sqrt_KS(fmpz_poly_t b, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrt_divconquer(fmpz * res, const fmpz * poly, slong len, int exact) noexcept
    int fmpz_poly_sqrt_divconquer(fmpz_poly_t b, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrt(fmpz * res, const fmpz * poly, slong len) noexcept
    int fmpz_poly_sqrt(fmpz_poly_t b, const fmpz_poly_t a) noexcept
    int _fmpz_poly_sqrt_series(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    int fmpz_poly_sqrt_series(fmpz_poly_t b, const fmpz_poly_t a, slong n) noexcept
    void _fmpz_poly_power_sums_naive(fmpz * res, const fmpz * poly, slong len, slong n) noexcept
    void fmpz_poly_power_sums_naive(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void fmpz_poly_power_sums(fmpz_poly_t res, const fmpz_poly_t poly, slong n) noexcept
    void _fmpz_poly_power_sums_to_poly(fmpz * res, const fmpz * poly, slong len) noexcept
    void fmpz_poly_power_sums_to_poly(fmpz_poly_t res, const fmpz_poly_t Q) noexcept
    void _fmpz_poly_signature(slong * r1, slong * r2, const fmpz * poly, slong len) noexcept
    void fmpz_poly_signature(slong * r1, slong * r2, const fmpz_poly_t poly) noexcept
    void fmpz_poly_hensel_build_tree(slong * link, fmpz_poly_t * v, fmpz_poly_t * w, const nmod_poly_factor_t fac) noexcept
    void fmpz_poly_hensel_lift(fmpz_poly_t G, fmpz_poly_t H, fmpz_poly_t A, fmpz_poly_t B, const fmpz_poly_t f, const fmpz_poly_t g, const fmpz_poly_t h, const fmpz_poly_t a, const fmpz_poly_t b, const fmpz_t p, const fmpz_t p1) noexcept
    void fmpz_poly_hensel_lift_without_inverse(fmpz_poly_t Gout, fmpz_poly_t Hout, const fmpz_poly_t f, const fmpz_poly_t g, const fmpz_poly_t h, const fmpz_poly_t a, const fmpz_poly_t b, const fmpz_t p, const fmpz_t p1) noexcept
    void fmpz_poly_hensel_lift_only_inverse(fmpz_poly_t Aout, fmpz_poly_t Bout, const fmpz_poly_t G, const fmpz_poly_t H, const fmpz_poly_t a, const fmpz_poly_t b, const fmpz_t p, const fmpz_t p1) noexcept
    void fmpz_poly_hensel_lift_tree_recursive(slong * link, fmpz_poly_t * v, fmpz_poly_t * w, fmpz_poly_t f, slong j, slong inv, const fmpz_t p0, const fmpz_t p1) noexcept
    void fmpz_poly_hensel_lift_tree(slong * link, fmpz_poly_t * v, fmpz_poly_t * w, fmpz_poly_t f, slong r, const fmpz_t p, slong e0, slong e1, slong inv) noexcept
    slong _fmpz_poly_hensel_start_lift(fmpz_poly_factor_t lifted_fac, slong * link, fmpz_poly_t * v, fmpz_poly_t * w, const fmpz_poly_t f, const nmod_poly_factor_t local_fac, slong N) noexcept
    slong _fmpz_poly_hensel_continue_lift(fmpz_poly_factor_t lifted_fac, slong * link, fmpz_poly_t * v, fmpz_poly_t * w, const fmpz_poly_t f, slong prev, slong curr, slong N, const fmpz_t p) noexcept
    void fmpz_poly_hensel_lift_once(fmpz_poly_factor_t lifted_fac, const fmpz_poly_t f, const nmod_poly_factor_t local_fac, slong N) noexcept
    int _fmpz_poly_print(const fmpz * poly, slong len) noexcept
    int fmpz_poly_print(const fmpz_poly_t poly) noexcept
    int _fmpz_poly_print_pretty(const fmpz * poly, slong len, const char * x) noexcept
    int fmpz_poly_print_pretty(const fmpz_poly_t poly, const char * x) noexcept
    int _fmpz_poly_fprint(FILE * file, const fmpz * poly, slong len) noexcept
    int fmpz_poly_fprint(FILE * file, const fmpz_poly_t poly) noexcept
    int _fmpz_poly_fprint_pretty(FILE * file, const fmpz * poly, slong len, const char * x) noexcept
    int fmpz_poly_fprint_pretty(FILE * file, const fmpz_poly_t poly, const char * x) noexcept
    int fmpz_poly_read(fmpz_poly_t poly) noexcept
    int fmpz_poly_read_pretty(fmpz_poly_t poly, char **x) noexcept
    int fmpz_poly_fread(FILE * file, fmpz_poly_t poly) noexcept
    int fmpz_poly_fread_pretty(FILE * file, fmpz_poly_t poly, char **x) noexcept
    void fmpz_poly_get_nmod_poly(nmod_poly_t Amod, const fmpz_poly_t A) noexcept
    void fmpz_poly_set_nmod_poly(fmpz_poly_t A, const nmod_poly_t Amod) noexcept
    void fmpz_poly_set_nmod_poly_unsigned(fmpz_poly_t A, const nmod_poly_t Amod) noexcept
    void _fmpz_poly_CRT_ui_precomp(fmpz * res, const fmpz * poly1, slong len1, const fmpz_t m1, mp_srcptr poly2, slong len2, mp_limb_t m2, mp_limb_t m2inv, fmpz_t m1m2, mp_limb_t c, int sign) noexcept
    void _fmpz_poly_CRT_ui(fmpz * res, const fmpz * poly1, slong len1, const fmpz_t m1, mp_srcptr poly2, slong len2, mp_limb_t m2, mp_limb_t m2inv, int sign) noexcept
    void fmpz_poly_CRT_ui(fmpz_poly_t res, const fmpz_poly_t poly1, const fmpz_t m, const nmod_poly_t poly2, int sign) noexcept
    void _fmpz_poly_product_roots_fmpz_vec(fmpz * poly, const fmpz * xs, slong n) noexcept
    void fmpz_poly_product_roots_fmpz_vec(fmpz_poly_t poly, const fmpz * xs, slong n) noexcept
    void _fmpz_poly_product_roots_fmpq_vec(fmpz * poly, const fmpq * xs, slong n) noexcept
    void fmpz_poly_product_roots_fmpq_vec(fmpz_poly_t poly, const fmpq * xs, slong n) noexcept
    void _fmpz_poly_bound_roots(fmpz_t bound, const fmpz * poly, slong len) noexcept
    void fmpz_poly_bound_roots(fmpz_t bound, const fmpz_poly_t poly) noexcept
    void _fmpz_poly_num_real_roots_sturm(slong * n_neg, slong * n_pos, const fmpz * pol, slong len) noexcept
    slong fmpz_poly_num_real_roots_sturm(const fmpz_poly_t pol) noexcept
    slong _fmpz_poly_num_real_roots(const fmpz * pol, slong len) noexcept
    slong fmpz_poly_num_real_roots(const fmpz_poly_t pol) noexcept
    void _fmpz_poly_cyclotomic(fmpz * a, ulong n, mp_ptr factors, slong num_factors, ulong phi) noexcept
    void fmpz_poly_cyclotomic(fmpz_poly_t poly, ulong n) noexcept
    ulong _fmpz_poly_is_cyclotomic(const fmpz * poly, slong len) noexcept
    ulong fmpz_poly_is_cyclotomic(const fmpz_poly_t poly) noexcept
    void _fmpz_poly_cos_minpoly(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_cos_minpoly(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_swinnerton_dyer(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_swinnerton_dyer(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_chebyshev_t(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_chebyshev_t(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_chebyshev_u(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_chebyshev_u(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_legendre_pt(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_legendre_pt(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_hermite_h(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_hermite_h(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_hermite_he(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_hermite_he(fmpz_poly_t poly, ulong n) noexcept
    void _fmpz_poly_fibonacci(fmpz * coeffs, ulong n) noexcept
    void fmpz_poly_fibonacci(fmpz_poly_t poly, ulong n) noexcept
    void arith_eulerian_polynomial(fmpz_poly_t res, ulong n) noexcept
    void _fmpz_poly_eta_qexp(fmpz * f, slong r, slong len) noexcept
    void fmpz_poly_eta_qexp(fmpz_poly_t f, slong r, slong n) noexcept
    void _fmpz_poly_theta_qexp(fmpz * f, slong r, slong len) noexcept
    void fmpz_poly_theta_qexp(fmpz_poly_t f, slong r, slong n) noexcept
    void fmpz_poly_CLD_bound(fmpz_t res, const fmpz_poly_t f, slong n) noexcept

from .fmpz_poly_macros cimport *



# sage_setup: distribution = sagemath-objects
# distutils: language = c++
# distutils: libraries = gmpxx gmp

from sage.libs.gmp.types cimport mpz_t, mpq_t

cdef extern from 'gmpxx.h':
    cdef cppclass mpz_class:
        mpz_class()
        mpz_class(int i)
        mpz_class(mpz_t z)
        mpz_class(mpz_class)
        mpz_t get_mpz_t()
        mpz_class operator % (mpz_class, mpz_class)

    cdef cppclass mpq_class:
        mpq_class()
        mpz_t get_num_mpz_t()
        mpz_t get_den_mpz_t()



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/fmpz.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    fmpz PTR_TO_COEFF(__mpz_struct * ptr) noexcept
    __mpz_struct * COEFF_TO_PTR(fmpz f) noexcept
    __mpz_struct * _fmpz_new_mpz() noexcept
    void _fmpz_clear_mpz(fmpz f) noexcept
    void _fmpz_cleanup_mpz_content() noexcept
    void _fmpz_cleanup() noexcept
    __mpz_struct * _fmpz_promote(fmpz_t f) noexcept
    __mpz_struct * _fmpz_promote_val(fmpz_t f) noexcept
    void _fmpz_demote(fmpz_t f) noexcept
    void _fmpz_demote_val(fmpz_t f) noexcept
    bint _fmpz_is_canonical(const fmpz_t f) noexcept
    void fmpz_init(fmpz_t f) noexcept
    void fmpz_init2(fmpz_t f, ulong limbs) noexcept
    void fmpz_clear(fmpz_t f) noexcept
    void fmpz_init_set(fmpz_t f, const fmpz_t g) noexcept
    void fmpz_init_set_ui(fmpz_t f, ulong g) noexcept
    void fmpz_init_set_si(fmpz_t f, slong g) noexcept
    void fmpz_randbits(fmpz_t f, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_randtest(fmpz_t f, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_randtest_unsigned(fmpz_t f, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_randtest_not_zero(fmpz_t f, flint_rand_t state, flint_bitcnt_t bits) noexcept
    void fmpz_randm(fmpz_t f, flint_rand_t state, const fmpz_t m) noexcept
    void fmpz_randtest_mod(fmpz_t f, flint_rand_t state, const fmpz_t m) noexcept
    void fmpz_randtest_mod_signed(fmpz_t f, flint_rand_t state, const fmpz_t m) noexcept
    void fmpz_randprime(fmpz_t f, flint_rand_t state, flint_bitcnt_t bits, int proved) noexcept
    slong fmpz_get_si(const fmpz_t f) noexcept
    ulong fmpz_get_ui(const fmpz_t f) noexcept
    void fmpz_get_uiui(mp_limb_t * hi, mp_limb_t * low, const fmpz_t f) noexcept
    mp_limb_t fmpz_get_nmod(const fmpz_t f, nmod_t mod) noexcept
    double fmpz_get_d(const fmpz_t f) noexcept
    void fmpz_set_mpf(fmpz_t f, const mpf_t x) noexcept
    void fmpz_get_mpf(mpf_t x, const fmpz_t f) noexcept
    void fmpz_get_mpfr(mpfr_t x, const fmpz_t f, mpfr_rnd_t rnd) noexcept
    double fmpz_get_d_2exp(slong * exp, const fmpz_t f) noexcept
    void fmpz_get_mpz(mpz_t x, const fmpz_t f) noexcept
    int fmpz_get_mpn(mp_ptr * n, fmpz_t n_in) noexcept
    char * fmpz_get_str(char * str, int b, const fmpz_t f) noexcept
    void fmpz_set_si(fmpz_t f, slong val) noexcept
    void fmpz_set_ui(fmpz_t f, ulong val) noexcept
    void fmpz_set_d(fmpz_t f, double c) noexcept
    void fmpz_set_d_2exp(fmpz_t f, double d, slong exp) noexcept
    void fmpz_neg_ui(fmpz_t f, ulong val) noexcept
    void fmpz_set_uiui(fmpz_t f, mp_limb_t hi, mp_limb_t lo) noexcept
    void fmpz_neg_uiui(fmpz_t f, mp_limb_t hi, mp_limb_t lo) noexcept
    void fmpz_set_signed_uiui(fmpz_t f, ulong hi, ulong lo) noexcept
    void fmpz_set_signed_uiuiui(fmpz_t f, ulong hi, ulong mid, ulong lo) noexcept
    void fmpz_set_ui_array(fmpz_t out, const ulong * input, slong n) noexcept
    void fmpz_set_signed_ui_array(fmpz_t out, const ulong * input, slong n) noexcept
    void fmpz_get_ui_array(ulong * out, slong n, const fmpz_t input) noexcept
    void fmpz_get_signed_ui_array(ulong * out, slong n, const fmpz_t input) noexcept
    void fmpz_get_signed_uiui(ulong * hi, ulong * lo, const fmpz_t input) noexcept
    void fmpz_set_mpz(fmpz_t f, const mpz_t x) noexcept
    int fmpz_set_str(fmpz_t f, const char * str, int b) noexcept
    void fmpz_set_ui_smod(fmpz_t f, mp_limb_t x, mp_limb_t m) noexcept
    void flint_mpz_init_set_readonly(mpz_t z, const fmpz_t f) noexcept
    void flint_mpz_clear_readonly(mpz_t z) noexcept
    void fmpz_init_set_readonly(fmpz_t f, const mpz_t z) noexcept
    void fmpz_clear_readonly(fmpz_t f) noexcept
    int fmpz_read(fmpz_t f) noexcept
    int fmpz_fread(FILE * file, fmpz_t f) noexcept
    size_t fmpz_inp_raw(fmpz_t x, FILE * fin) noexcept
    int fmpz_print(const fmpz_t x) noexcept
    int fmpz_fprint(FILE * file, const fmpz_t x) noexcept
    size_t fmpz_out_raw(FILE * fout, const fmpz_t x ) noexcept
    size_t fmpz_sizeinbase(const fmpz_t f, int b) noexcept
    flint_bitcnt_t fmpz_bits(const fmpz_t f) noexcept
    mp_size_t fmpz_size(const fmpz_t f) noexcept
    int fmpz_sgn(const fmpz_t f) noexcept
    flint_bitcnt_t fmpz_val2(const fmpz_t f) noexcept
    void fmpz_swap(fmpz_t f, fmpz_t g) noexcept
    void fmpz_set(fmpz_t f, const fmpz_t g) noexcept
    void fmpz_zero(fmpz_t f) noexcept
    void fmpz_one(fmpz_t f) noexcept
    int fmpz_abs_fits_ui(const fmpz_t f) noexcept
    int fmpz_fits_si(const fmpz_t f) noexcept
    void fmpz_setbit(fmpz_t f, ulong i) noexcept
    int fmpz_tstbit(const fmpz_t f, ulong i) noexcept
    mp_limb_t fmpz_abs_lbound_ui_2exp(slong * exp, const fmpz_t x, int bits) noexcept
    mp_limb_t fmpz_abs_ubound_ui_2exp(slong * exp, const fmpz_t x, int bits) noexcept
    int fmpz_cmp(const fmpz_t f, const fmpz_t g) noexcept
    int fmpz_cmp_ui(const fmpz_t f, ulong g) noexcept
    int fmpz_cmp_si(const fmpz_t f, slong g) noexcept
    int fmpz_cmpabs(const fmpz_t f, const fmpz_t g) noexcept
    int fmpz_cmp2abs(const fmpz_t f, const fmpz_t g) noexcept
    bint fmpz_equal(const fmpz_t f, const fmpz_t g) noexcept
    bint fmpz_equal_ui(const fmpz_t f, ulong g) noexcept
    bint fmpz_equal_si(const fmpz_t f, slong g) noexcept
    bint fmpz_is_zero(const fmpz_t f) noexcept
    bint fmpz_is_one(const fmpz_t f) noexcept
    bint fmpz_is_pm1(const fmpz_t f) noexcept
    bint fmpz_is_even(const fmpz_t f) noexcept
    bint fmpz_is_odd(const fmpz_t f) noexcept
    void fmpz_neg(fmpz_t f1, const fmpz_t f2) noexcept
    void fmpz_abs(fmpz_t f1, const fmpz_t f2) noexcept
    void fmpz_add(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_add_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_add_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_sub(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_sub_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_sub_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_mul(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_mul_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_mul_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_mul2_uiui(fmpz_t f, const fmpz_t g, ulong x, ulong y) noexcept
    void fmpz_mul_2exp(fmpz_t f, const fmpz_t g, ulong e) noexcept
    void fmpz_one_2exp(fmpz_t f, ulong e) noexcept
    void fmpz_addmul(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_addmul_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_addmul_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_submul(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_submul_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_submul_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_fmma(fmpz_t f, const fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_t d) noexcept
    void fmpz_fmms(fmpz_t f, const fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_t d) noexcept
    void fmpz_cdiv_qr(fmpz_t f, fmpz_t s, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_fdiv_qr(fmpz_t f, fmpz_t s, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_tdiv_qr(fmpz_t f, fmpz_t s, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_ndiv_qr(fmpz_t f, fmpz_t s, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_cdiv_q(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_fdiv_q(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_tdiv_q(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_cdiv_q_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_fdiv_q_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_tdiv_q_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_cdiv_q_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_fdiv_q_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_tdiv_q_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_cdiv_q_2exp(fmpz_t f, const fmpz_t g, ulong exp) noexcept
    void fmpz_fdiv_q_2exp(fmpz_t f, const fmpz_t g, ulong exp) noexcept
    void fmpz_tdiv_q_2exp(fmpz_t f, const fmpz_t g, ulong exp) noexcept
    void fmpz_fdiv_r(fmpz_t s, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_cdiv_r_2exp(fmpz_t s, const fmpz_t g, ulong exp) noexcept
    void fmpz_fdiv_r_2exp(fmpz_t s, const fmpz_t g, ulong exp) noexcept
    void fmpz_tdiv_r_2exp(fmpz_t s, const fmpz_t g, ulong exp) noexcept
    ulong fmpz_cdiv_ui(const fmpz_t g, ulong h) noexcept
    ulong fmpz_fdiv_ui(const fmpz_t g, ulong h) noexcept
    ulong fmpz_tdiv_ui(const fmpz_t g, ulong h) noexcept
    void fmpz_divexact(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_divexact_si(fmpz_t f, const fmpz_t g, slong h) noexcept
    void fmpz_divexact_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_divexact2_uiui(fmpz_t f, const fmpz_t g, ulong x, ulong y) noexcept
    int fmpz_divisible(const fmpz_t f, const fmpz_t g) noexcept
    int fmpz_divisible_si(const fmpz_t f, slong g) noexcept
    int fmpz_divides(fmpz_t q, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_mod(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    ulong fmpz_mod_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_smod(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_preinvn_init(fmpz_preinvn_t inv, const fmpz_t f) noexcept
    void fmpz_preinvn_clear(fmpz_preinvn_t inv) noexcept
    void fmpz_fdiv_qr_preinvn(fmpz_t f, fmpz_t s, const fmpz_t g, const fmpz_t h, const fmpz_preinvn_t hinv) noexcept
    void fmpz_pow_ui(fmpz_t f, const fmpz_t g, ulong x) noexcept
    void fmpz_ui_pow_ui(fmpz_t f, ulong g, ulong x) noexcept
    int fmpz_pow_fmpz(fmpz_t f, const fmpz_t g, const fmpz_t x) noexcept
    void fmpz_powm_ui(fmpz_t f, const fmpz_t g, ulong e, const fmpz_t m) noexcept
    void fmpz_powm(fmpz_t f, const fmpz_t g, const fmpz_t e, const fmpz_t m) noexcept
    slong fmpz_clog(const fmpz_t x, const fmpz_t b) noexcept
    slong fmpz_clog_ui(const fmpz_t x, ulong b) noexcept
    slong fmpz_flog(const fmpz_t x, const fmpz_t b) noexcept
    slong fmpz_flog_ui(const fmpz_t x, ulong b) noexcept
    double fmpz_dlog(const fmpz_t x) noexcept
    int fmpz_sqrtmod(fmpz_t b, const fmpz_t a, const fmpz_t p) noexcept
    void fmpz_sqrt(fmpz_t f, const fmpz_t g) noexcept
    void fmpz_sqrtrem(fmpz_t f, fmpz_t r, const fmpz_t g) noexcept
    bint fmpz_is_square(const fmpz_t f) noexcept
    int fmpz_root(fmpz_t r, const fmpz_t f, slong n) noexcept
    bint fmpz_is_perfect_power(fmpz_t root, const fmpz_t f) noexcept
    void fmpz_fac_ui(fmpz_t f, ulong n) noexcept
    void fmpz_fib_ui(fmpz_t f, ulong n) noexcept
    void fmpz_bin_uiui(fmpz_t f, ulong n, ulong k) noexcept
    void _fmpz_rfac_ui(fmpz_t r, const fmpz_t x, ulong a, ulong b) noexcept
    void fmpz_rfac_ui(fmpz_t r, const fmpz_t x, ulong k) noexcept
    void fmpz_rfac_uiui(fmpz_t r, ulong x, ulong k) noexcept
    void fmpz_mul_tdiv_q_2exp(fmpz_t f, const fmpz_t g, const fmpz_t h, ulong exp) noexcept
    void fmpz_mul_si_tdiv_q_2exp(fmpz_t f, const fmpz_t g, slong x, ulong exp) noexcept
    void fmpz_gcd_ui(fmpz_t f, const fmpz_t g, ulong h) noexcept
    void fmpz_gcd(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_gcd3(fmpz_t f, const fmpz_t a, const fmpz_t b, const fmpz_t c) noexcept
    void fmpz_lcm(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_gcdinv(fmpz_t d, fmpz_t a, const fmpz_t f, const fmpz_t g) noexcept
    void fmpz_xgcd(fmpz_t d, fmpz_t a, fmpz_t b, const fmpz_t f, const fmpz_t g) noexcept
    void fmpz_xgcd_canonical_bezout(fmpz_t d, fmpz_t a, fmpz_t b, const fmpz_t f, const fmpz_t g) noexcept
    void fmpz_xgcd_partial(fmpz_t co2, fmpz_t co1, fmpz_t r2, fmpz_t r1, const fmpz_t L) noexcept
    slong _fmpz_remove(fmpz_t x, const fmpz_t f, double finv) noexcept
    slong fmpz_remove(fmpz_t rop, const fmpz_t op, const fmpz_t f) noexcept
    int fmpz_invmod(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    void fmpz_negmod(fmpz_t f, const fmpz_t g, const fmpz_t h) noexcept
    int fmpz_jacobi(const fmpz_t a, const fmpz_t n) noexcept
    int fmpz_kronecker(const fmpz_t a, const fmpz_t n) noexcept
    void fmpz_divides_mod_list(fmpz_t xstart, fmpz_t xstride, fmpz_t xlength, const fmpz_t a, const fmpz_t b, const fmpz_t n) noexcept
    int fmpz_bit_pack(mp_limb_t * arr, flint_bitcnt_t shift, flint_bitcnt_t bits, const fmpz_t coeff, int negate, int borrow) noexcept
    int fmpz_bit_unpack(fmpz_t coeff, mp_limb_t * arr, flint_bitcnt_t shift, flint_bitcnt_t bits, int negate, int borrow) noexcept
    void fmpz_bit_unpack_unsigned(fmpz_t coeff, const mp_limb_t * arr, flint_bitcnt_t shift, flint_bitcnt_t bits) noexcept
    void fmpz_complement(fmpz_t r, const fmpz_t f) noexcept
    void fmpz_clrbit(fmpz_t f, ulong i) noexcept
    void fmpz_combit(fmpz_t f, ulong i) noexcept
    void fmpz_and(fmpz_t r, const fmpz_t a, const fmpz_t b) noexcept
    void fmpz_or(fmpz_t r, const fmpz_t a, const fmpz_t b) noexcept
    void fmpz_xor(fmpz_t r, const fmpz_t a, const fmpz_t b) noexcept
    ulong fmpz_popcnt(const fmpz_t a) noexcept
    void fmpz_CRT_ui(fmpz_t out, const fmpz_t r1, const fmpz_t m1, ulong r2, ulong m2, int sign) noexcept
    void fmpz_CRT(fmpz_t out, const fmpz_t r1, const fmpz_t m1, const fmpz_t r2, const fmpz_t m2, int sign) noexcept
    void fmpz_multi_mod_ui(mp_limb_t * out, const fmpz_t input, const fmpz_comb_t comb, fmpz_comb_temp_t temp) noexcept
    void fmpz_multi_CRT_ui(fmpz_t output, mp_srcptr residues, const fmpz_comb_t comb, fmpz_comb_temp_t ctemp, int sign) noexcept
    void fmpz_comb_init(fmpz_comb_t comb, mp_srcptr primes, slong num_primes) noexcept
    void fmpz_comb_temp_init(fmpz_comb_temp_t temp, const fmpz_comb_t comb) noexcept
    void fmpz_comb_clear(fmpz_comb_t comb) noexcept
    void fmpz_comb_temp_clear(fmpz_comb_temp_t temp) noexcept
    void fmpz_multi_CRT_init(fmpz_multi_CRT_t CRT) noexcept
    int fmpz_multi_CRT_precompute(fmpz_multi_CRT_t CRT, const fmpz * moduli, slong len) noexcept
    void fmpz_multi_CRT_precomp(fmpz_t output, const fmpz_multi_CRT_t P, const fmpz * inputs, int sign) noexcept
    int fmpz_multi_CRT(fmpz_t output, const fmpz * moduli, const fmpz * values, slong len, int sign) noexcept
    void fmpz_multi_CRT_clear(fmpz_multi_CRT_t P) noexcept
    bint fmpz_is_strong_probabprime(const fmpz_t n, const fmpz_t a) noexcept
    bint fmpz_is_probabprime_lucas(const fmpz_t n) noexcept
    bint fmpz_is_probabprime_BPSW(const fmpz_t n) noexcept
    bint fmpz_is_probabprime(const fmpz_t p) noexcept
    bint fmpz_is_prime_pseudosquare(const fmpz_t n) noexcept
    bint fmpz_is_prime_pocklington(fmpz_t F, fmpz_t R, const fmpz_t n, mp_ptr pm1, slong num_pm1) noexcept
    void _fmpz_nm1_trial_factors(const fmpz_t n, mp_ptr pm1, slong * num_pm1, ulong limit) noexcept
    bint fmpz_is_prime_morrison(fmpz_t F, fmpz_t R, const fmpz_t n, mp_ptr pp1, slong num_pp1) noexcept
    void _fmpz_np1_trial_factors(const fmpz_t n, mp_ptr pp1, slong * num_pp1, ulong limit) noexcept
    bint fmpz_is_prime(const fmpz_t n) noexcept
    void fmpz_lucas_chain(fmpz_t Vm, fmpz_t Vm1, const fmpz_t A, const fmpz_t m, const fmpz_t n) noexcept
    void fmpz_lucas_chain_full(fmpz_t Vm, fmpz_t Vm1, const fmpz_t A, const fmpz_t B, const fmpz_t m, const fmpz_t n) noexcept
    void fmpz_lucas_chain_double(fmpz_t U2m, fmpz_t U2m1, const fmpz_t Um, const fmpz_t Um1, const fmpz_t A, const fmpz_t B, const fmpz_t n) noexcept
    void fmpz_lucas_chain_add(fmpz_t Umn, fmpz_t Umn1, const fmpz_t Um, const fmpz_t Um1, const fmpz_t Un, const fmpz_t Un1, const fmpz_t A, const fmpz_t B, const fmpz_t n) noexcept
    void fmpz_lucas_chain_mul(fmpz_t Ukm, fmpz_t Ukm1, const fmpz_t Um, const fmpz_t Um1, const fmpz_t A, const fmpz_t B, const fmpz_t k, const fmpz_t n) noexcept
    void fmpz_lucas_chain_VtoU(fmpz_t Um, fmpz_t Um1, const fmpz_t Vm, const fmpz_t Vm1, const fmpz_t A, const fmpz_t B, const fmpz_t Dinv, const fmpz_t n) noexcept
    int fmpz_divisor_in_residue_class_lenstra(fmpz_t fac, const fmpz_t n, const fmpz_t r, const fmpz_t s) noexcept
    void fmpz_nextprime(fmpz_t res, const fmpz_t n, int proved) noexcept
    void fmpz_primorial(fmpz_t res, ulong n) noexcept
    void fmpz_factor_euler_phi(fmpz_t res, const fmpz_factor_t fac) noexcept
    void fmpz_euler_phi(fmpz_t res, const fmpz_t n) noexcept
    int fmpz_factor_moebius_mu(const fmpz_factor_t fac) noexcept
    int fmpz_moebius_mu(const fmpz_t n) noexcept
    void fmpz_factor_divisor_sigma(fmpz_t res, ulong k, const fmpz_factor_t fac) noexcept
    void fmpz_divisor_sigma(fmpz_t res, ulong k, const fmpz_t n) noexcept

from .fmpz_macros cimport *



# sage_setup: distribution = sagemath-objects
from sage.structure.element cimport Element
from sage.categories.map cimport Map


cdef class Morphism(Map):
    pass

cdef class SetMorphism(Morphism):
    cdef object _function
    cpdef bint _eq_c_impl(left, Element right) noexcept

cdef class SetIsomorphism(SetMorphism):
    cdef object _inverse



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Decoders

Representation of an error-correction algorithm for a code.

AUTHORS:

- David Joyner (2009-02-01): initial version

- David Lucas (2015-06-29): abstract class version
"""
#*****************************************************************************
#       Copyright (C) 2009 David Joyner <wdjoyner@gmail.com>
#                     2015 David Lucas <david.lucas@inria.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your preference).
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.structure.sage_object import SageObject


class Decoder(SageObject):
    r"""
    Abstract top-class for :class:`Decoder` objects.

    Every decoder class for linear codes (of any metric) should inherit from
    this abstract class.

    To implement a decoder, you need to:

    - inherit from :class:`Decoder`

    - call ``Decoder.__init__`` in the subclass constructor.
      Example: ``super().__init__(code, input_space, connected_encoder_name)``.
      By doing that, your subclass will have all the parameters described above initialized.

    - Then, you need to override one of decoding methods, either :meth:`decode_to_code` or
      :meth:`decode_to_message`. You can also override the optional method :meth:`decoding_radius`.

    - By default, comparison of :class:`Decoder` (using methods ``__eq__`` and ``__ne__`` ) are
      by memory reference: if you build the same decoder twice, they will be different. If you
      need something more clever, override ``__eq__`` and ``__ne__`` in your subclass.

    - As :class:`Decoder` is not designed to be instantiated, it does not have any representation
      methods. You should implement ``_repr_`` and ``_latex_`` methods in the subclass.
    """

    @classmethod
    def decoder_type(cls):
        r"""
        Return the set of types of ``self``.

        This method can be called on both an uninstantiated decoder class,
        or on an instance of a decoder class.

        The types of a decoder are a set of labels commonly associated with
        decoders which describe the nature and behaviour of the decoding
        algorithm. It should be considered as an informal descriptor but
        can be coarsely relied upon for e.g. program logic.

        The following are the most common types and a brief definition:

        ======================  ================================================
        Decoder type            Definition
        ======================  ================================================
        always-succeed          The decoder always returns a closest codeword if
                                the number of errors is up to the decoding
                                radius.
        bounded-distance        Any vector with Hamming distance at most
                                ``decoding_radius()`` to a codeword is
                                decodable to some codeword. If ``might-fail`` is
                                also a type, then this is not a guarantee but an
                                expectancy.
        complete                The decoder decodes every word in the ambient
                                space of the code.
        dynamic                 Some of the decoder's types will only be
                                determined at construction time
                                (depends on the parameters).
        half-minimum-distance   The decoder corrects up to half the minimum
                                distance, or a specific lower bound thereof.
        hard-decision           The decoder uses no information on which
                                positions are more likely to be in error or not.
        list-decoder            The decoder outputs a list of likely codewords,
                                instead of just a single codeword.
        might-fail              The decoder can fail at decoding even within its
                                usual promises, e.g. bounded distance.
        not-always-closest      The decoder does not guarantee to always return a
                                closest codeword.
        probabilistic           The decoder has internal randomness which can affect
                                running time and the decoding result.
        soft-decision           As part of the input, the decoder takes
                                reliability information on which positions are
                                more likely to be in error. Such a decoder only
                                works for specific channels.
        ======================  ================================================

        EXAMPLES:

        We call it on a class::

            sage: codes.decoders.LinearCodeSyndromeDecoder.decoder_type()
            {'dynamic', 'hard-decision'}

        We can also call it on a instance of a :class:`Decoder` class::

            sage: G = Matrix(GF(2), [[1, 0, 0, 1], [0, 1, 1, 1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.decoder_type()
            {'complete', 'hard-decision', 'might-error'}
        """
        return cls._decoder_type

    def _instance_decoder_type(self):
        r"""
        See the documentation of :meth:`decoder_type`.

        EXAMPLES:

        Test to satisfy the doc-testing framework::

            sage: G = Matrix(GF(2), [[1, 0, 0, 1], [0, 1, 1, 1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.decoder_type() #indirect doctest
            {'complete', 'hard-decision', 'might-error'}
        """
        return self._decoder_type

    def __init__(self, code, input_space, connected_encoder_name):
        r"""
        Initialize mandatory parameters for :class:`Decoder` objects.

        This method only exists for inheritance purposes as it initializes
        parameters that need to be known by every decoder. An abstract
        decoder object should never be created.

        INPUT:

        - ``code`` -- the associated code of ``self``

        - ``input_space`` -- the input space of ``self``, which is the ambient space
          of ``self``'s ``code``

        - ``connected_encoder_name`` -- the associated encoder, which will be
          used by ``self`` to recover elements from the message space

        EXAMPLES:

        We first create a new :class:`Decoder` subclass::

            sage: from sage.coding.decoder import Decoder
            sage: class DecoderExample(Decoder):
            ....:   def __init__(self, code):
            ....:       in_space = code.ambient_space()
            ....:       connected_enc = "GeneratorMatrix"
            ....:       super().__init__(code, in_space, connected_enc)

        We now create a member of our brand new class::

            sage: G = Matrix(GF(2), [[1, 0, 0, 1], [0, 1, 1, 1]])
            sage: C = LinearCode(G)
            sage: D = DecoderExample(C)

        We can check its parameters::

            sage: D.input_space()
            Vector space of dimension 4 over Finite Field of size 2
            sage: D.connected_encoder()
            Generator matrix-based encoder for [4, 2] linear code over GF(2)
            sage: D.code()
            [4, 2] linear code over GF(2)
        """
        self.decoder_type = self._instance_decoder_type
        self._code = code
        self._input_space = input_space
        self._connected_encoder_name = connected_encoder_name

    def __hash__(self):
        r"""
        Return the hash value of ``self``.

        This is a generic implementation which should be overwritten on decoders
        with extra arguments.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0],[1,0,0,1,1,0,0],[0,1,0,1,0,1,0],[1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: hash(D) #random
            7575380076354998465
        """
        C = self.code()
        Str = str(C)
        return hash((C, Str)) ^ hash(Str) ^ hash(C)

    def __ne__(self, other):
        r"""
        Test inequality of ``self`` and ``other``.

        This is a generic implementation, which returns the inverse of
        ``__eq__`` for ``self``.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0],[1,0,0,1,1,0,0],[0,1,0,1,0,1,0],[1,1,0,1,0,0,1]])
            sage: D1 = LinearCode(G).decoder()
            sage: D2 = LinearCode(G).decoder()
            sage: D1 != D2
            False
            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0],[1,0,0,1,1,0,0],[0,1,0,1,0,1,0],[1,1,0,1,0,1,1]])
            sage: D2 = LinearCode(G).decoder()
            sage: D1 != D2
            True
        """
        return not self == other

    def decode_to_code(self, r):
        r"""
        Correct the errors in ``r`` and return a codeword.

        This is a default implementation which assumes that the method
        :meth:`decode_to_message` has been implemented, else it returns an exception.

        INPUT:

        - ``r`` -- a element of the input space of ``self``

        OUTPUT: a vector of :meth:`code`

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector(GF(2), (1, 1, 0, 0, 1, 1, 0))
            sage: word in C
            True
            sage: w_err = word + vector(GF(2), (1, 0, 0, 0, 0, 0, 0))
            sage: w_err in C
            False
            sage: D = C.decoder()
            sage: D.decode_to_code(w_err)
            (1, 1, 0, 0, 1, 1, 0)
        """
        if hasattr(self, "defaulting_decode_to_message"):
            raise NotImplementedError
        else:
            word = self.decode_to_message(r)
            return self.connected_encoder().encode(word)

    def connected_encoder(self):
        r"""
        Return the connected encoder of ``self``.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.connected_encoder()
            Generator matrix-based encoder for [7, 4] linear code over GF(2)
        """
        return self.code().encoder(encoder_name=self._connected_encoder_name)

    def decode_to_message(self, r):
        r"""
        Decode ``r`` to the message space of :meth:`connected_encoder`.

        This is a default implementation, which assumes that the
        method :meth:`decode_to_code` has been implemented, else it
        returns an exception.

        INPUT:

        - ``r`` -- a element of the input space of ``self``

        OUTPUT: a vector of :meth:`message_space`

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector(GF(2), (1, 1, 0, 0, 1, 1, 0))
            sage: w_err = word + vector(GF(2), (1, 0, 0, 0, 0, 0, 0))
            sage: D = C.decoder()
            sage: D.decode_to_message(w_err)
            (1, 1, 0, 0)
        """
        self.defaulting_decode_to_message = True
        return self.code().unencode(self.decode_to_code(r))

    def code(self):
        r"""
        Return the code for this :class:`Decoder`.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.code()
            [7, 4] linear code over GF(2)
        """
        return self._code

    def message_space(self):
        r"""
        Return the message space of ``self``'s :meth:`connected_encoder`.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.message_space()
            Vector space of dimension 4 over Finite Field of size 2
        """
        return self.connected_encoder().message_space()

    def input_space(self):
        r"""
        Return the input space of ``self``.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = C.decoder()
            sage: D.input_space()
            Vector space of dimension 7 over Finite Field of size 2
        """
        if hasattr(self, "_input_space"):
            return self._input_space
        else:
            raise NotImplementedError("Decoder does not have an _input_space parameter")

    @abstract_method(optional=True)
    def decoding_radius(self, **kwargs):
        r"""
        Return the maximal number of errors that ``self`` is able to correct.

        This is an abstract method and it should be implemented in subclasses.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: D = codes.decoders.LinearCodeSyndromeDecoder(C)
            sage: D.decoding_radius()
            1
        """
        raise NotImplementedError


class DecodingError(Exception):
    r"""
    Special exception class to indicate an error during decoding.
    """
    pass



# sage_setup: distribution = sagemath-flint
"""
Saturation over ZZ
"""

from copy import copy

from sage.arith.misc import binomial, GCD as gcd
from sage.matrix import matrix_integer_dense_hnf
from sage.matrix.constructor import identity_matrix, random_matrix
from sage.misc.randstate import current_randstate
from sage.misc.verbose import verbose
from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF
from sage.rings.integer_ring import ZZ


def p_saturation(A, p, proof=True):
    """
    INPUT:

    - ``A`` -- a matrix over ZZ
    - ``p`` -- a prime
    - ``proof`` -- boolean (default: ``True``)

    OUTPUT:

    The p-saturation of the matrix A, i.e., a new matrix in Hermite form
    whose row span a ZZ-module that is p-saturated.

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_dense_saturation import p_saturation
        sage: A = matrix(ZZ, 2, 2, [3,2,3,4]); B = matrix(ZZ, 2,3,[1,2,3,4,5,6])
        sage: A.det()
        6
        sage: C = A*B; C
        [11 16 21]
        [19 26 33]
        sage: C2 = p_saturation(C, 2); C2
        [ 1  8 15]
        [ 0  9 18]
        sage: C2.index_in_saturation()
        9
        sage: C3 = p_saturation(C, 3); C3
        [ 1  0 -1]
        [ 0  2  4]
        sage: C3.index_in_saturation()
        2
    """
    tm = verbose("%s-saturating a %sx%s matrix" % (p, A.nrows(), A.ncols()))
    H = A.hermite_form(include_zero_rows=False, proof=proof)
    while True:
        if p == 2:
            A = H.change_ring(GF(p))
        else:
            try:
                # Faster than change_ring
                A = H._reduce(p)
            except (OverflowError, ImportError):
                # fall back to generic GF(p) matrices
                A = H.change_ring(GF(p))
        assert A.nrows() <= A.ncols()
        K = A.kernel()
        if K.dimension() == 0:
            verbose("done saturating", tm)
            return H
        B = K.basis_matrix().lift()
        C = ((B * H) / p).change_ring(ZZ)
        H = H.stack(C).hermite_form(include_zero_rows=False, proof=proof)
    verbose("done saturating", tm)


def random_sublist_of_size(k, n):
    """
    INPUT:

    - ``k`` -- integer
    - ``n`` -- integer

    OUTPUT: a randomly chosen sublist of ``range(k)`` of size `n`

    EXAMPLES::

        sage: import sage.matrix.matrix_integer_dense_saturation as s
        sage: l = s.random_sublist_of_size(10, 3)
        sage: len(l)
        3
        sage: l_check = [-1] + l + [10]
        sage: all(l_check[i] < l_check[i+1] for i in range(4))
        True
        sage: l = s.random_sublist_of_size(10, 7)
        sage: len(l)
        7
        sage: l_check = [-1] + l + [10]
        sage: all(l_check[i] < l_check[i+1] for i in range(8))
        True
    """
    if n > k:
        raise ValueError("n must be <= len(v)")
    if n == k:
        return list(range(k))
    if n >= k // 2 + 5:
        # use complement
        w = random_sublist_of_size(k, k - n)
        m = set(w)
        w = [z for z in range(k) if z not in m]
        assert len(w) == n
        return w

    randrange = current_randstate().python_random().randrange

    w = set()
    while len(w) < n:
        z = randrange(k)
        if z not in w:
            w.add(z)
    return sorted(w)


def solve_system_with_difficult_last_row(B, A):
    """
    Solve the matrix equation ``B*Z = A`` when the last row of `B`
    contains huge entries.

    INPUT:

    - ``B`` -- a square n x n nonsingular matrix with painful big bottom row
    - ``A`` -- an n x k matrix

    OUTPUT: the unique solution to ``B*Z = As``

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_dense_saturation import solve_system_with_difficult_last_row
        sage: B = matrix(ZZ, 3, [1,2,3, 3,-1,2,939239082,39202803080,2939028038402834]); A = matrix(ZZ,3,2,[1,2,4,3,-1,0])
        sage: X = solve_system_with_difficult_last_row(B, A); X
        [  290668794698843/226075992027744         468068726971/409557956572]
        [-226078357385539/1582531944194208       1228691305937/2866905696004]
        [      2365357795/1582531944194208           -17436221/2866905696004]
        sage: B*X == A
        True
    """
    # See the comments in the function of the same name in matrix_integer_dense_hnf.py.
    # This function is just a generalization of that one to A a matrix.
    C = copy(B)
    while True:
        C[C.nrows()-1] = random_matrix(ZZ,1,C.ncols()).row(0)
        try:
            X = C.solve_right(A)
        except ValueError:
            verbose("Try difficult solve again with different random vector")
        else:
            break
    D = B.matrix_from_rows(range(C.nrows()-1))
    N = D._rational_kernel_flint()
    if N.ncols() != 1:
        verbose("Difficult solve quickly failed.  Using direct approach.")
        return B.solve_right(A)

    tm = verbose("Recover correct linear combinations")
    k = N.matrix_from_columns([0])

    # The sought for solution Z to B*Z = A is some linear combination
    #       Z = X + alpha*k
    #  Let w be the last row of B; then Z satisfies
    #       w * Z = A'
    # where A' is the last row of A.  Thus
    #       w * (X + alpha*k) = A'
    # so    w * X + alpha*w*k = A'
    # so    alpha*w*k  = A' - w*X.
    w = B[-1]  # last row of B
    A_prime = A[-1]  # last row of A
    lhs = w*k
    rhs = A_prime - w * X

    if lhs[0] == 0:
        verbose("Difficult solve quickly failed.  Using direct approach.")
        return B.solve_right(A)

    for i in range(X.ncols()):
        alpha = rhs[i] / lhs[0]
        X.set_column(i, (X.matrix_from_columns([i]) + alpha*k).list())
    verbose("Done getting linear combinations.", tm)
    return X


def saturation(A, proof=True, p=0, max_dets=5):
    r"""
    Compute a saturation matrix of `A`.

    INPUT:

    - ``A`` -- a matrix over `\ZZ`
    - ``proof`` -- boolean (default: ``True``)
    - ``p`` -- integer (default: 0); if not 0 only guarantees that output is
      `p`-saturated
    - ``max_dets`` -- integer (default: 4); max number of dets of submatrices to
      compute

    OUTPUT: matrix; saturation of the matrix `A`

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_dense_saturation import saturation
        sage: A = matrix(ZZ, 2, 2, [3,2,3,4]); B = matrix(ZZ, 2,3,[1,2,3,4,5,6]); C = A*B
        sage: C
        [11 16 21]
        [19 26 33]
        sage: C.index_in_saturation()
        18
        sage: S = saturation(C); S
        [11 16 21]
        [-2 -3 -4]
        sage: S.index_in_saturation()
        1
        sage: saturation(C, proof=False)
        [11 16 21]
        [-2 -3 -4]
        sage: saturation(C, p=2)
        [11 16 21]
        [-2 -3 -4]
        sage: saturation(C, p=2, max_dets=1)
        [11 16 21]
        [-2 -3 -4]
    """
    # Find a submatrix of full rank and instead saturate that matrix.
    r = A.rank()
    if A.is_square() and r == A.nrows():
        return identity_matrix(ZZ, r)
    if A.nrows() > r:
        P = []
        while len(P) < r:
            P = matrix_integer_dense_hnf.probable_pivot_rows(A)
        A = A.matrix_from_rows(P)

    # Factor out all common factors from all rows, just in case.
    A = copy(A)
    A._factor_out_common_factors_from_each_row()

    if A.nrows() <= 1:
        return A

    A, zero_cols = A._delete_zero_columns()

    if max_dets > 0:
        # Take the GCD of at most num_dets randomly chosen determinants.
        nr = A.nrows()
        nc = A.ncols()
        d = 0
        trials = min(binomial(nc, nr), max_dets)
        already_tried = []
        while len(already_tried) < trials:
            v = random_sublist_of_size(nc, nr)
            tm = verbose('saturation -- checking det condition on submatrix')
            d = gcd(d, A.matrix_from_columns(v).determinant(proof=proof))
            verbose('saturation -- got det down to %s' % d, tm)
            if gcd(d, p) == 1:
                return A._insert_zero_columns(zero_cols)
            already_tried.append(v)

        if gcd(d, p) == 1:
            # already p-saturated
            return A._insert_zero_columns(zero_cols)

        # Factor and p-saturate at each p.
        # This is not a good algorithm, because all the HNF's in it are really slow!
        #
        #tm = verbose('factoring gcd %s of determinants'%d)
        #limit = 2**31-1
        #F = d.factor(limit = limit)
        #D = [p for p, e in F if p <= limit]
        #B = [n for n, e in F if n > limit]  # all big factors -- there will only be at most one
        #assert len(B) <= 1
        #C = B[0]
        #for p in D:
        #    A = p_saturation(A, p=p, proof=proof)

    # This is a really simple but powerful algorithm.
    # FACT: If A is a matrix of full rank, then hnf(transpose(A))^(-1)*A is a saturation of A.
    # To make this practical we use solve_system_with_difficult_last_row, since the
    # last column of HNF's are typically the only really big ones.
    B = A.transpose().hermite_form(include_zero_rows=False, proof=proof)
    B = B.transpose()

    # Now compute B^(-1) * A
    C = solve_system_with_difficult_last_row(B, A)
    return C.change_ring(ZZ)._insert_zero_columns(zero_cols)


def index_in_saturation(A, proof=True):
    r"""
    The index of A in its saturation.

    INPUT:

    - ``A`` -- matrix over `\ZZ`

    - ``proof`` -- boolean (``True`` or ``False``)

    OUTPUT: integer

    EXAMPLES::

        sage: from sage.matrix.matrix_integer_dense_saturation import index_in_saturation
        sage: A = matrix(ZZ, 2, 2, [3,2,3,4]); B = matrix(ZZ, 2,3,[1,2,3,4,5,6]); C = A*B; C
        [11 16 21]
        [19 26 33]
        sage: index_in_saturation(C)
        18
        sage: W = C.row_space()
        sage: S = W.saturation()
        sage: W.index_in(S)
        18

    For any zero matrix the index in its saturation is 1 (see :issue:`13034`)::

        sage: m = matrix(ZZ, 3)
        sage: m
        [0 0 0]
        [0 0 0]
        [0 0 0]
        sage: m.index_in_saturation()
        1
        sage: m = matrix(ZZ, 2, 3)
        sage: m
        [0 0 0]
        [0 0 0]
        sage: m.index_in_saturation()
        1

    TESTS::

        sage: zero = matrix(ZZ, [[]])
        sage: zero.index_in_saturation()
        1
    """
    r = A.rank()
    if r == 0:
        return ZZ.one()
    if r < A.nrows():
        A = A.hermite_form(proof=proof, include_zero_rows=False)
    if A.is_square():
        return abs(A.determinant(proof=proof))
    A = A.transpose()
    A = A.hermite_form(proof=proof, include_zero_rows=False)
    return abs(A.determinant(proof=proof))



# sage_setup: distribution = sagemath-schemes
r"""
Scheme implementation overview

Various parts of schemes were implemented by different authors. This document
aims to give an overview of the different classes of schemes working together
coherently.

Generic
-------

- **Scheme:**
  A scheme whose datatype might not be defined in terms of algebraic equations:
  e.g. the Jacobian of a curve may be represented by means of a Scheme.

- **AlgebraicScheme:**
  A scheme defined by means of polynomial equations, which may be reducible or
  defined over a ring other than a field.  In particular, the defining ideal
  need not be a radical ideal, and an algebraic scheme may be defined over
  `\mathrm{Spec}(R)`.

- **AmbientSpaces:**
  Most effective models of algebraic scheme will be defined not by generic
  gluings, but by embeddings in some fixed ambient space.

Ambients
--------

- **AffineSpace:**
  Affine spaces and their affine subschemes form the most important universal
  objects from which algebraic schemes are built.  The affine spaces form
  universal objects in the sense that a morphism is uniquely determined by the
  images of its coordinate functions and any such images determine a
  well-defined morphism.

  By default affine spaces will embed in some ordinary projective space, unless
  it is created as an affine patch of another object.

- **ProjectiveSpace:**
  Projective spaces are the most natural ambient spaces for most projective
  objects.  They are locally universal objects.

- **ProjectiveSpace_ordinary (not implemented):**
  The ordinary projective spaces have the standard weights `[1,..,1]` on their
  coefficients.

- **ProjectiveSpace_weighted (not implemented):**
  A special subtype for non-standard weights.

- **ToricVariety:**
  Toric varieties are (partial) compactifications of algebraic tori `(\CC^*)^n`
  compatible with torus action. Affine and projective spaces are examples of
  toric varieties, but it is not envisioned that these special cases should
  inherit from :class:`~sage.schemes.toric.variety.ToricVariety`.

Subschemes
----------

- **AlgebraicScheme_subscheme_affine:**
  An algebraic scheme defined by means of an embedding in a
  fixed ambient affine space.

- **AlgebraicScheme_subscheme_projective:**
  An algebraic scheme defined by means of an embedding in a fixed ambient
  projective space.

- **QuasiAffineScheme (not yet implemented):**
  An open subset `U = X \setminus Z` of a closed subset `X` of affine space;
  note that this is mathematically a quasi-projective scheme, but its
  ambient space is an affine space and its points are represented by
  affine rather than projective points.

  .. NOTE::

    AlgebraicScheme_quasi is implemented, as a base class for this.

- **QuasiProjectiveScheme (not yet implemented):**
  An open subset of a closed subset of projective space; this datatype
  stores the defining polynomial, polynomials, or ideal defining the
  projective closure `X` plus the closed subscheme `Z` of `X` whose complement
  `U = X \setminus Z` is the quasi-projective scheme.

  .. NOTE::

    The quasi-affine and quasi-projective datatype lets one create schemes
    like the multiplicative group scheme
    `\mathbb{G}_m = \mathbb{A}^1\setminus \{(0)\}`
    and the non-affine scheme `\mathbb{A}^2\setminus \{(0,0)\}`.  The latter
    is not affine and is not of the form `\mathrm{Spec}(R)`.

Point sets
----------

- **PointSets and points over a ring (to do):**
  For algebraic schemes `X/S` and `T/S` over `S`, one can form the point set
  `X(T)` of morphisms from `T\to X` over `S`.

  A projective space object in the category of schemes is a locally free object
  -- the images of the generator functions *locally* determine a point.  Over a
  field, one can choose one of the standard affine patches by the condition
  that a coordinate function `X_i \ne 0`.

  ::

        sage: PP.<X,Y,Z> = ProjectiveSpace(2, QQ)
        sage: PP
        Projective Space of dimension 2 over Rational Field
        sage: PP(QQ)
        Set of rational points of Projective Space
        of dimension 2 over Rational Field
        sage: PP(QQ)([-2, 3, 5])
        (-2/5 : 3/5 : 1)

  Over a ring, this is not true anymore. For example, even over an integral
  domain which is not a PID, there may be no *single* affine patch which covers
  a point.

  ::

        sage: R.<x> = ZZ[]
        sage: S.<t> = R.quo(x^2 + 5)
        sage: P.<X,Y,Z> = ProjectiveSpace(2, S)
        sage: P(S)
        Set of rational points of Projective Space of dimension 2 over
        Univariate Quotient Polynomial Ring in t over Integer Ring with
        modulus x^2 + 5

  In order to represent the projective point `(2:1+t) = (1-t:3)` we note that
  the first representative is not well-defined at the prime `p = (2,1+t)` and
  the second element is not well-defined at the prime `q = (1-t,3)`, but that
  `p + q = (1)`, so globally the pair of coordinate representatives is
  well-defined.

  ::

        sage: P([2, 1 + t])
        (2 : t + 1 : 1)

  In fact, we need a test ``R.ideal([2, 1 + t]) == R.ideal([1])`` in order to make
  this meaningful.

Berkovich Analytic Spaces
-------------------------

- **Berkovich Analytic Space (not yet implemented)**
  The construction of analytic spaces from schemes due to Berkovich. Any Berkovich
  space should inherit from :class:`Berkovich`

- **Berkovich Analytic Space over Cp**
  A special case of the general Berkovich analytic space construction.
  Affine Berkovich space over `\CC_p` is the set of seminorms on the polynomial
  ring `\CC_p[x]`, while projective Berkovich space over `\CC_p` is the one-point compactification
  of affine Berkovich space `\CC_p`. Points are represented using the classification (due
  to Berkovich) of a corresponding decreasing sequence of disks in `\CC_p`.

AUTHORS:

- David Kohel, William Stein (2006-01-03): initial version

- Andrey Novoseltsev (2010-09-24): updated due to addition of toric varieties
"""



# sage_setup: distribution = sagemath-gap
"""
Short tests for GAP
"""

from sage.libs.gap.libgap import libgap
from sage.misc.temporary_file import tmp_filename


def test_write_to_file():
    """
    Test that libgap can write to files.

    See :issue:`16502`, :issue:`15833`.

    EXAMPLES::

        sage: from sage.libs.gap.test import test_write_to_file
        sage: test_write_to_file()
    """
    fname = tmp_filename()
    message = "Ceci n'est pas une groupe"
    libgap.PrintTo(fname, message)
    with open(fname) as f:
        assert f.read() == message
    SystemFile = libgap.function_factory('StringFile')
    assert SystemFile(fname).sage() == message
